#define _GNU_SOURCE
#include <pthread.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/syscall.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <signal.h>
#include <ctype.h>
#include <sys/prctl.h>
#include <unistd.h>

#include "binder.h"

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static struct binder_state *bs;

static void hexdump_(void *_data, size_t byte_count) {
  printf("hexdump(%p, 0x%lx)\n", _data, (unsigned long)byte_count);
  for (unsigned long byte_offset = 0; byte_offset < byte_count; byte_offset += 16) {
    unsigned char *bytes = ((unsigned char*)_data) + byte_offset;
    unsigned long line_bytes = (byte_count - byte_offset > 16) ?
            16 : (byte_count - byte_offset);
    char line[1000];
    char *linep = line;
    linep += sprintf(linep, "%08lx  ", byte_offset);
    for (int i=0; i<16; i++) {
      if (i >= line_bytes) {
        linep += sprintf(linep, "   ");
      } else {
        linep += sprintf(linep, "%02hhx ", bytes[i]);
      }
    }
    linep += sprintf(linep, " |");
    for (int i=0; i<line_bytes; i++) {
      if (isalnum(bytes[i]) || ispunct(bytes[i]) || bytes[i] == ' ') {
        *(linep++) = bytes[i];
      } else {
        *(linep++) = '.';
      }
    }
    linep += sprintf(linep, "|");
    puts(line);
  }
}

int manager_binder_handler(struct binder_state *bs, struct binder_transaction_data *txn, struct binder_io *msg, struct binder_io *reply) {
  printf("got transaction!\n");
  bio_put_uint32(reply, 0);
  return 0;
}

static int mm_holder_fn(void *dummy) {
  SYSCHK(prctl(PR_SET_PDEATHSIG, SIGKILL));
  if (getppid() == 1) exit(0);
  SYSCHK(mmap((void*)0x200000, 0x400000, PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_PRIVATE, -1, 0));
  SYSCHK(madvise((void*)0x200000, 0x400000, MADV_HUGEPAGE));
  *(volatile char*)0x200000;
  printf("remapped as THP in mm_holder\n");

  // keep the mmap lock occupied as much as possible
  while (1) munmap((void*)0x1000, 0x1000);
}

int main(int argc, char** argv) {
  setbuf(stdout, NULL);

  if (argc == 2 && strcmp(argv[1], "REEXEC") == 0) {
    bs = binder_open_fromfd(101, 0x400000);
    if (!bs) errx(1, "binder_open_fromfd()");

    // The ping allocates binder memory in both the parent and the child.
    printf("### PING\n");
    struct binder_io msg;
    struct binder_io reply;
    char data[0x1000];
    uint32_t handle = 0;

    bio_init(&msg, data, sizeof(data), 4);
    bio_put_uint32(&msg, 0);
    binder_call(bs, &msg, &reply, 0, 1);
    // Query is deallocated, reply isn't deallocated yet - collect page from
    // the parent process.
    binder_done(bs, &msg, &reply);
    printf("### PING DONE\n");

    sleep(5);
    pid_t parent = getppid();
    if (parent != 1)
      kill(parent, SIGKILL);

    char *thp_test = SYSCHK(mmap((void*)0x600000, 0x200000, PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_PRIVATE, -1, 0));
    SYSCHK(madvise(thp_test, 0x200000, MADV_HUGEPAGE));
    hexdump_(thp_test, 0x1000);

    return 0;
  }

  pid_t parent = getpid();
  pid_t child = SYSCHK(fork());
  if (child == 0) {
    SYSCHK(prctl(PR_SET_PDEATHSIG, SIGKILL));
    if (getppid() != parent) exit(0);

    sleep(1);
    int binder_fd = SYSCHK(open("/dev/binder", O_RDWR));
    SYSCHK(dup2(binder_fd, 101));

    char *mm_holder_stack = SYSCHK(mmap(NULL, 1024*1024*8, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, -1, 0));
    SYSCHK(clone(mm_holder_fn, mm_holder_stack+1024*1024*8, CLONE_VM|SIGCHLD, NULL));
    /*
    pid_t mm_holder = SYSCHK(syscall(__NR_clone, CLONE_VM|SIGCHLD, mm_holder_stack+1024*1024*8, NULL, NULL, 0));
    if (mm_holder == 0) {
    }
    */

    sleep(1);
    SYSCHK(execl("/proc/self/exe", argv[0], "REEXEC", NULL));
  } else {
    bs = binder_open(0x400000);
    if (bs == NULL) err(1, "binder_open");
    if (binder_become_context_manager(bs)) err(1, "binder_become_context_manager");

    binder_loop(bs, manager_binder_handler);
  
    return 0;
  }
}
