#include <pthread.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>

#include <fcntl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/prctl.h>
#include <unistd.h>

#include "binder.h"

static struct binder_state *bs;

int manager_binder_handler(struct binder_state *bs, struct binder_transaction_data *txn, struct binder_io *msg, struct binder_io *reply) {
  printf("got transaction!\n");
  bio_put_uint32(reply, 0);
  sleep(2);
  return 0;
}

void *munmap_thread(void *arg) {
  sleep(1);
  printf("$$$ calling munmap now...\n");
  prctl(PR_SET_NAME, "SLOW1");
  munmap(bs->mapped, bs->mapsize);
  prctl(PR_SET_NAME, "dummy");
  printf("$$$ munmap done\n");
  return NULL;
}

int main(int argc, char** argv) {
  setbuf(stdout, NULL);

  pid_t parent = getpid();
  pid_t child = fork();
  if (child == -1) err(1, "fork");
  if (child == 0) {
    if (prctl(PR_SET_PDEATHSIG, SIGKILL)) err(1, "prctl");
    if (getppid() != parent) exit(0);

    bs = binder_open(0x400000);
    if (bs == NULL) err(1, "binder_open");

    pthread_t thread;
    if (pthread_create(&thread, NULL, munmap_thread, NULL))
      errx(1, "pthread_create");

    // The ping allocates binder memory in both the parent and the child.
    printf("### PING\n");
    struct binder_io msg;
    struct binder_io reply;
    char data[0x1000];
    uint32_t handle = 0;

    bio_init(&msg, data, sizeof(data), 4);
    bio_put_uint32(&msg, 0);
    binder_call(bs, &msg, &reply, 0, 1);
    binder_done(bs, &msg, &reply);
    printf("### PING DONE\n");


    sleep(5);
    kill(parent, SIGKILL);
  } else {
    bs = binder_open(0x400000);
    if (bs == NULL) err(1, "binder_open");
    if (binder_become_context_manager(bs)) err(1, "binder_become_context_manager");

    prctl(PR_SET_NAME, "SLOW2");
    binder_loop(bs, manager_binder_handler);
  
    return 0;
  }
}
